/**
 * @author T. Przedzinski
 * @date 19 Nov 2013
 *
 * @brief Corrects energy of the particles based on 
 *        generated_mass() information from HepMC::GenEvent
 *
 * Introduces class FourVector for manipulating momentum
 * of the particles stored in HepMC file.
 *
 * Included by hepmc-tauola-redecay.cxx but the function call
 * is commented out by default
 * This algorithm may create severe problems if kinematical impurities of
 * the events are not the consequence of rounding errors only 
 */
#include "HepMC/GenParticle.h"
#include "HepMC/GenEvent.h"
#include <vector>
#include <iostream>
using namespace std;

/** Threshold above which the correction is considered too large */
const double CORRECTION_THRESHOLD = 0.0001;

/** Class for storing information on particle momentum, including mass */
class FourVector
{
public:
  FourVector();
  FourVector(double _x, double _y, double _z, double _e, double _m);
  FourVector(const HepMC::FourVector& x);
  FourVector(const FourVector& x);
  FourVector operator+(const FourVector& y);
  FourVector operator-(const FourVector& y);
  FourVector operator+=(const FourVector& x);
  FourVector operator-=(const FourVector& x);
  double recalculated_mass();
  void print();
public:
  double px,py,pz,e,m;
};

FourVector::FourVector():px(0.),py(0.),pz(0.),e(0.),m(0.)                                                         {}
FourVector::FourVector(double _x, double _y, double _z, double _e, double _m=0.):px(_x),py(_y),pz(_z),e(_e),m(_m) {}
FourVector::FourVector(const HepMC::FourVector& x):px(x.px()),py(x.py()),pz(x.pz()),e(x.e()),m(0.)                {}
FourVector::FourVector(const FourVector& x):px(x.px),py(x.py),pz(x.pz),e(x.e),m(x.m)                              {}

FourVector FourVector::operator+(const FourVector& y)
{
  return FourVector(px+y.px,py+y.py,pz+y.pz,e+y.e,0.);
}

FourVector FourVector::operator-(const FourVector& y)
{
  return FourVector(px-y.px,py-y.py,pz-y.pz,e-y.e,0.);
}

FourVector FourVector::operator+=(const FourVector& x)
{
  return (*this)=(*this)+x;
}

FourVector FourVector::operator-=(const FourVector& x)
{
  return (*this)=(*this)-x;
}

double FourVector::recalculated_mass()
{
  double mm = e*e-px*px-py*py-pz*pz;
  if(mm>0.0)      return  sqrt(mm);
  else if(mm<0.0) return -sqrt(-mm);
  else return 0.0;
}

void FourVector::print()
{
  printf("%+13.6e %+13.6e %+13.6e %+13.6e m: %+13.6e recalculated_m: %+13.6e\n",
           px,    py,     pz,     e,         m,                      recalculated_mass() );
}

/*******************************************************************************
  Get daughters of HepMC::GenParticle

  Recursively searches for final-state daughters of 'x'
*******************************************************************************/
inline vector<HepMC::GenParticle*> getDaughters(HepMC::GenParticle *x)
{
  vector<HepMC::GenParticle*> daughters;
  if(!x->end_vertex()) return daughters;

  // Check decay products of 'x'
  for(HepMC::GenVertex::particles_out_const_iterator p = x->end_vertex()->particles_out_const_begin(); p!=x->end_vertex()->particles_out_const_end(); ++p)
  {
    HepMC::GenParticle *pp = *p;
    HepMC::FourVector   mm = pp->momentum();

    // If the daughter of 'x' has its end vertex - recursively read
    // all of its daughters. Ignore decays of pi0 if present
    if( pp->end_vertex() && pp->pdg_id()!=111)
    {
      vector<HepMC::GenParticle*> sub_daughters = getDaughters(pp);
      daughters.insert(daughters.end(),sub_daughters.begin(),sub_daughters.end());
    }
    // Otherwise - add this particle to the list of daughters.
    else daughters.push_back(pp);
  }

  return daughters;
}

/*****************************************************************************
  Correct event
  Performs following operation:
  - for each stable tau decay product its energy is recalculated from the mass
    and momentum
  - the four-momentum of tau is recalculated from the sum of four-momenta
    of its decay products
  - checks if resulting operation did not have introduced sizeable
    modifications. If it does, warning message is printed. This may indicate
    problems in the production file other than rounding error
*****************************************************************************/
void CorrectEvent(HepMC::GenEvent *evt)
{
  HepMC::GenParticle *pX    = NULL;
  HepMC::GenParticle *pTaum = NULL;
  HepMC::GenParticle *pTaup = NULL;
  
  // Get tau pair and intermediate boson
	for ( HepMC::GenEvent::particle_const_iterator p = evt->particles_begin();
	      p != evt->particles_end(); ++p )
	{
    if( (*p)->pdg_id() == 15 && (*p)->end_vertex() && abs((*(*p)->end_vertex()->particles_out_const_begin())->pdg_id())==16) pTaum = *p;
    if( (*p)->pdg_id() ==-15 && (*p)->end_vertex() && abs((*(*p)->end_vertex()->particles_out_const_begin())->pdg_id())==16) pTaup = *p;
	}

  if(!pTaum || !pTaup) return;
  
  // Get tau daughters
  vector<HepMC::GenParticle*> taupDaughters = getDaughters(pTaup);
  vector<HepMC::GenParticle*> taumDaughters = getDaughters(pTaum);

  FourVector taup_sum,taum_sum;

  // Correct tau+ daughters
  for(unsigned int i=0;i<taupDaughters.size();i++)
  {
    HepMC::FourVector m = taupDaughters[i]->momentum();
    FourVector buf = m;
    buf.m = taupDaughters[i]->generated_mass();
    
    // Correct energy
    buf.e = sqrt(buf.m*buf.m + buf.px*buf.px + buf.py*buf.py + buf.pz*buf.pz);
    
    m.setE(buf.e);
    
    taupDaughters[i]->set_momentum(m);
    
    taup_sum += buf;
  }
  
  // Correct tau- daughters
  for(unsigned int i=0;i<taumDaughters.size();i++)
  {
    HepMC::FourVector m = taumDaughters[i]->momentum();
    FourVector buf = m;
    buf.m = taumDaughters[i]->generated_mass();

    // Correct energy
    buf.e = sqrt(buf.m*buf.m + buf.px*buf.px + buf.py*buf.py + buf.pz*buf.pz);
    
    m.setE(buf.e);
    
    taumDaughters[i]->set_momentum(m);
    
    taum_sum += buf;
  }
  
  // Check the size of the corrections
  FourVector taup_diff = pTaup->momentum();  
  FourVector taum_diff = pTaum->momentum();
  
  taup_diff -= taup_sum;
  taum_diff -= taum_sum;
  
  double diff1 = sqrt(taup_diff.px*taup_diff.px + taup_diff.py*taup_diff.py + taup_diff.pz*taup_diff.pz + taup_diff.e*taup_diff.e);
  double diff2 = sqrt(taum_diff.px*taum_diff.px + taum_diff.py*taum_diff.py + taum_diff.pz*taum_diff.pz + taum_diff.e*taum_diff.e);
  
  cout<<diff1<<" "<<diff2<<endl;
  
  if( diff1>CORRECTION_THRESHOLD*taup_sum.e || diff2>CORRECTION_THRESHOLD*taum_sum.e )
  {
    cout<<"CorrectEvent: WARNING! Large correction: "<<diff1/taup_sum.e<<" "<<diff2/taum_sum.e<<endl;
  }
  
  // Correct tau momentum
  HepMC::FourVector fv_taup(taup_sum.px,taup_sum.py,taup_sum.pz,taup_sum.e);
  HepMC::FourVector fv_taum(taum_sum.px,taum_sum.py,taum_sum.pz,taum_sum.e);
  
  pTaup->set_momentum(fv_taup);
  pTaum->set_momentum(fv_taum);
}
